@@ -22,6 +22,7 @@ from tensorflow.python.framework import dtypes
 from tensorflow.python.framework import indexed_slices as indexed_slices_lib
 from tensorflow.python.framework import ops
 from tensorflow.python.framework import sparse_tensor
+from tensorflow.python.framework import tensor
 from tensorflow.python.framework import tensor_shape
 from tensorflow.python.framework import tensor_util
 from tensorflow.python.ops import array_ops
@@ -36,19 +37,21 @@ from tensorflow.python.ops import sparse_ops
 
 
 @ops.RegisterGradient("Pack")
-def _PackGrad(op, grad):
+def _PackGrad(op: ops.Operation, grad):
   """Gradient for pack op."""
   return array_ops_stack.unstack(
       grad, num=op.get_attr("N"), axis=op.get_attr("axis"))
 
 
 @ops.RegisterGradient("Unpack")
-def _UnpackGrad(op, *grads):
+def _UnpackGrad(op: ops.Operation, *grads):
   """Gradient for unpack op."""
   return array_ops_stack.stack(grads, axis=op.get_attr("axis"))
 
 
-def _ConcatGradHelper(op, grad, start_value_index, end_value_index, dim_index):
+def _ConcatGradHelper(
+    op: ops.Operation, grad, start_value_index, end_value_index, dim_index
+):
   """Gradient for concat op.
 
   Args:
@@ -91,7 +94,7 @@ def _ConcatGradHelper(op, grad, start_value_index, end_value_index, dim_index):
     for x in inputs:
       input_shape = array_ops.shape(x)
       if not isinstance(input_shape,
-                        ops.Tensor) or input_shape.op.type != "Const":
+                        tensor.Tensor) or input_shape.op.type != "Const":
         fully_known = False
         break
       sizes.append(input_shape)
@@ -109,7 +112,7 @@ def _ConcatGradHelper(op, grad, start_value_index, end_value_index, dim_index):
   input_values = op.inputs[start_value_index:end_value_index]
 
   out_grads = []
-  if isinstance(grad, ops.Tensor):
+  if isinstance(grad, tensor.Tensor):
     if context.executing_eagerly() or isinstance(concat_dim, ops.EagerTensor):
       # Using mod here for convenience since concat_dim is already verified
       # in concat implementation to be within the allowed [-rank, rank) range.
@@ -215,7 +218,7 @@ def _ConcatGradHelper(op, grad, start_value_index, end_value_index, dim_index):
 
 
 @ops.RegisterGradient("Concat")
-def _ConcatGrad(op, grad):
+def _ConcatGrad(op: ops.Operation, grad):
   return _ConcatGradHelper(
       op,
       grad,
@@ -225,7 +228,7 @@ def _ConcatGrad(op, grad):
 
 
 @ops.RegisterGradient("ConcatV2")
-def _ConcatGradV2(op, grad):
+def _ConcatGradV2(op: ops.Operation, grad):
   return _ConcatGradHelper(
       op, grad, start_value_index=0, end_value_index=-1, dim_index=-1)
 
@@ -234,7 +237,7 @@ ops.NotDifferentiable("ConcatOffset")
 
 
 @ops.RegisterGradient("Slice")
-def _SliceGrad(op, grad):
+def _SliceGrad(op: ops.Operation, grad):
   """Gradient for Slice op."""
   # Create an Nx2 padding where the first column represents how many
   # zeros are to be prepended for each dimension, and the second
@@ -264,7 +267,7 @@ def _SliceGrad(op, grad):
 
 
 @ops.RegisterGradient("StridedSlice")
-def _StridedSliceGrad(op, grad):
+def _StridedSliceGrad(op: ops.Operation, grad):
   """Gradient for StridedSlice op."""
   begin = op.inputs[1]
   end = op.inputs[2]
@@ -299,7 +302,7 @@ def _StridedSliceGrad(op, grad):
 
 
 @ops.RegisterGradient("StridedSliceGrad")
-def _StridedSliceGradGrad(op, grad):
+def _StridedSliceGradGrad(op: ops.Operation, grad):
   """Gradient for StridedSliceGrad op."""
   begin = op.inputs[1]
   end = op.inputs[2]
@@ -318,7 +321,7 @@ def _StridedSliceGradGrad(op, grad):
 
 
 @ops.RegisterGradient("TensorStridedSliceUpdate")
-def _TensorStridedSliceUpdateGrad(op, grad):  # pylint:disable=missing-function-docstring
+def _TensorStridedSliceUpdateGrad(op: ops.Operation, grad):  # pylint:disable=missing-function-docstring
   begin = op.inputs[1]
   end = op.inputs[2]
   strides = op.inputs[3]
@@ -353,12 +356,12 @@ def _TensorStridedSliceUpdateGrad(op, grad):  # pylint:disable=missing-function-
 
 
 @ops.RegisterGradient("Split")
-def _SplitGrad(op, *grads):
+def _SplitGrad(op: ops.Operation, *grads):
   return None, array_ops.concat(list(grads), op.inputs[0])
 
 
 @ops.RegisterGradient("SplitV")
-def _SplitVGrad(op, *grads):
+def _SplitVGrad(op: ops.Operation, *grads):
   returnval = array_ops.concat(list(grads), op.inputs[2])
   returnval = [returnval] + [
       None,
@@ -386,19 +389,19 @@ def _MatrixDiagGrad(_, grad):
 
 
 @ops.RegisterGradient("MatrixDiagV2")
-def _MatrixDiagV2Grad(op, grad):
+def _MatrixDiagV2Grad(op: ops.Operation, grad):
   return array_ops.matrix_diag_part(
       grad, k=op.inputs[1]), None, None, None, None
 
 
 @ops.RegisterGradient("MatrixDiagV3")
-def _MatrixDiagV3Grad(op, grad):
+def _MatrixDiagV3Grad(op: ops.Operation, grad):
   return array_ops.matrix_diag_part(
       grad, k=op.inputs[1], align=op.get_attr("align")), None, None, None, None
 
 
 @ops.RegisterGradient("MatrixDiagPart")
-def _MatrixDiagPartGrad(op, grad):
+def _MatrixDiagPartGrad(op: ops.Operation, grad):
   matrix_shape = op.inputs[0].get_shape()[-2:]
   if matrix_shape.is_fully_defined() and matrix_shape[0] == matrix_shape[1]:
     return array_ops.matrix_diag(grad)
@@ -407,7 +410,7 @@ def _MatrixDiagPartGrad(op, grad):
 
 
 @ops.RegisterGradient("MatrixDiagPartV2")
-def _MatrixDiagPartV2Grad(op, grad):
+def _MatrixDiagPartV2Grad(op: ops.Operation, grad):
   """Gradient for MatrixDiagPartV2."""
   matrix_shape = op.inputs[0].get_shape()[-2:]
   if matrix_shape.is_fully_defined():
@@ -422,7 +425,7 @@ def _MatrixDiagPartV2Grad(op, grad):
 
 
 @ops.RegisterGradient("MatrixDiagPartV3")
-def _MatrixDiagPartV3Grad(op, grad):
+def _MatrixDiagPartV3Grad(op: ops.Operation, grad):
   """Gradient for MatrixDiagPartV3."""
   matrix_shape = op.inputs[0].get_shape()[-2:]
   align = op.get_attr("align")
@@ -440,7 +443,7 @@ def _MatrixDiagPartV3Grad(op, grad):
 
 
 @ops.RegisterGradient("MatrixSetDiag")
-def _MatrixSetDiagGrad(op, grad):
+def _MatrixSetDiagGrad(op: ops.Operation, grad):
   """Gradient for MatrixSetDiag."""
   input_shape = op.inputs[0].get_shape().merge_with(grad.get_shape())
   diag_shape = op.inputs[1].get_shape()
@@ -463,7 +466,7 @@ def _MatrixSetDiagGrad(op, grad):
 
 
 @ops.RegisterGradient("MatrixSetDiagV2")
-def _MatrixSetDiagGradV2(op, grad):
+def _MatrixSetDiagGradV2(op: ops.Operation, grad):
   """Gradient for MatrixSetDiagV2."""
   diag_shape = op.inputs[1].get_shape()
   if not diag_shape.is_fully_defined():
@@ -499,7 +502,7 @@ def _MatrixSetDiagGradV2(op, grad):
 
 
 @ops.RegisterGradient("MatrixSetDiagV3")
-def _MatrixSetDiagGradV3(op, grad):
+def _MatrixSetDiagGradV3(op: ops.Operation, grad):
   """Gradient for MatrixSetDiagV3."""
   diag_shape = op.inputs[1].get_shape()
   align = op.get_attr("align")
@@ -539,7 +542,7 @@ def _MatrixSetDiagGradV3(op, grad):
 
 
 @ops.RegisterGradient("MatrixBandPart")
-def _MatrixBandPartGrad(op, grad):
+def _MatrixBandPartGrad(op: ops.Operation, grad):
   num_lower = op.inputs[1]
   num_upper = op.inputs[2]
   return (array_ops.matrix_band_part(grad, num_lower, num_upper), None, None)
@@ -559,7 +562,7 @@ ops.NotDifferentiable("OnesLike")
 
 
 @ops.RegisterGradient("PreventGradient")
-def _PreventGradientGrad(op, _):
+def _PreventGradientGrad(op: ops.Operation, _):
   raise LookupError("Gradient explicitly disabled. Reason: %s" %
                     op.get_attr("message"))
 
@@ -579,7 +582,7 @@ def _IndexedSlicesToTensorNoWarning(indexed_slices):
 
 
 @ops.RegisterGradient("Gather")
-def _GatherGrad(op, grad):
+def _GatherGrad(op: ops.Operation, grad):
   """Gradient for Gather op."""
   # params can be large, so colocate the shape calculation with it.
   params = op.inputs[0]
@@ -649,7 +652,7 @@ def _BatchGatherGrad(params_shape, values, indices, batch_dims,
 
 
 @ops.RegisterGradient("GatherV2")
-def _GatherV2Grad(op, grad):
+def _GatherV2Grad(op: ops.Operation, grad):
   """Gradient for GatherV2 op."""
   # params can be large, so colocate the shape calculation with it.
   #
@@ -727,11 +730,18 @@ def _GatherV2Grad(op, grad):
     ], 0)
     params_grad = array_ops.transpose(params_grad, invert_transpose_dims)
 
+  if not isinstance(params_grad, indexed_slices_lib.IndexedSlices):
+    # Prevents mismatches in shapes when some tensor dimensions are zero.
+    params_grad = array_ops.reshape(
+        params_grad,
+        array_ops.shape(params)
+    )
+
   return [params_grad, None, None]
 
 
 @ops.RegisterGradient("GatherNd")
-def _GatherNdGrad(op, grad):
+def _GatherNdGrad(op: ops.Operation, grad):
   ref = op.inputs[0]
   indices = op.inputs[1]
   ref_shape = array_ops.shape(ref, out_type=indices.dtype)
@@ -744,7 +754,7 @@ def _GatherNdGrad(op, grad):
 
 
 @ops.RegisterGradient("ResourceGatherNd")
-def _ResourceGatherNdGrad(op, grad):  # pylint: disable=missing-docstring
+def _ResourceGatherNdGrad(op: ops.Operation, grad):  # pylint: disable=missing-docstring
   ref = op.inputs[0]
   indices = op.inputs[1]
   ref_shape = gen_resource_variable_ops.variable_shape(ref, indices.dtype)
@@ -757,7 +767,7 @@ def _ResourceGatherNdGrad(op, grad):  # pylint: disable=missing-docstring
 
 
 @ops.RegisterGradient("CheckNumerics")
-def _CheckNumericsGrad(op, grad):
+def _CheckNumericsGrad(op: ops.Operation, grad):
   """Gradient for check_numerics op."""
   return array_ops.check_numerics(
       grad,
@@ -766,7 +776,7 @@ def _CheckNumericsGrad(op, grad):
 
 
 @ops.RegisterGradient("CheckNumericsV2")
-def _CheckNumericsV2Grad(op, grad):
+def _CheckNumericsV2Grad(op: ops.Operation, grad):
   """Gradient for check_numerics op."""
   return array_ops.check_numerics_v2(
       grad,
@@ -800,7 +810,28 @@ ops.NotDifferentiable("StopGradient")
 
 
 @ops.RegisterGradient("Reshape")
-def _ReshapeGrad(op, grad):
+def _ReshapeGrad(op: ops.Operation, grad):
+  """Defines the gradient for `array_ops.reshape()`."""
+  input_shape = op.inputs[0].shape
+  if input_shape.rank is not None and not input_shape.is_fully_defined():
+    # If only one dimension is undefined, we can use a wildcard dimension in
+    # the argument to `reshape()` to avoid creating a data dependency via
+    # a dynamic `shape()` operation.
+    input_shape_as_list = input_shape.as_list()
+    undefined_dims = []
+    has_zero_dim = False
+    for i, dim in enumerate(input_shape_as_list):
+      if dim is None:
+        undefined_dims.append(i)
+      elif dim == 0:
+        # When the tensor has zero elements, the wildcard dimension
+        # is underspecified, and `reshape()` will arbitrarily set the unknown
+        # to `1`, triggering shape errors downstream.
+        has_zero_dim = True
+    if len(undefined_dims) == 1 and not has_zero_dim:
+      input_shape_as_list[undefined_dims[0]] = -1
+      return [array_ops.reshape(_IndexedSlicesToTensorNoWarning(grad),
+                                input_shape_as_list), None]
   return [
       array_ops.reshape(
           _IndexedSlicesToTensorNoWarning(grad), array_ops.shape(op.inputs[0])),
@@ -811,31 +842,31 @@ def _ReshapeGrad(op, grad):
 ops.NotDifferentiable("InvertPermutation")
 
 
-def _ReshapeToInput(op, grad):
+def _ReshapeToInput(op: ops.Operation, grad):
   """Reshapes the gradient to the shape of the original input."""
   return array_ops.reshape(
       _IndexedSlicesToTensorNoWarning(grad), array_ops.shape(op.inputs[0]))
 
 
 @ops.RegisterGradient("ExpandDims")
-def _ExpandDimsGrad(op, grad):
+def _ExpandDimsGrad(op: ops.Operation, grad):
   return [_ReshapeToInput(op, grad), None]
 
 
 @ops.RegisterGradient("Squeeze")
-def _SqueezeGrad(op, grad):
+def _SqueezeGrad(op: ops.Operation, grad):
   return _ReshapeToInput(op, grad)
 
 
 @ops.RegisterGradient("Transpose")
-def _TransposeGrad(op, grad):
+def _TransposeGrad(op: ops.Operation, grad):
   """Returns unshuffle(grad)."""
   p = op.inputs[1]
   return [array_ops.transpose(grad, array_ops.invert_permutation(p)), None]
 
 
 @ops.RegisterGradient("ConjugateTranspose")
-def _ConjugateTransposeGrad(op, grad):
+def _ConjugateTransposeGrad(op: ops.Operation, grad):
   """Returns conj(unshuffle(grad))."""
   p = op.inputs[1]
   return [
@@ -854,7 +885,7 @@ ops.NotDifferentiable("Size")
 
 
 @ops.RegisterGradient("Tile")
-def _TileGrad(op, grad):
+def _TileGrad(op: ops.Operation, grad):
   """Sum reduces grad along the tiled dimensions."""
   input_shape = array_ops.shape(op.inputs[0], out_type=op.inputs[1].dtype)
   # We interleave multiples and input_shape to get split_shape,
@@ -885,7 +916,7 @@ def _TileGrad(op, grad):
 ops.NotDifferentiable("BroadcastGradientArgs")
 
 
-def _PadGrad(op, grad):
+def _PadGrad(op: ops.Operation, grad):
   """Gradient for Pad."""
   # Pad introduces values around the original tensor, so the gradient function
   # slices the original shape out of the gradient."""
@@ -910,7 +941,7 @@ ops.RegisterGradient("PadV2")(_PadGrad)
 
 # ReverseSequence is just a permutation.  The gradient permutes back.
 @ops.RegisterGradient("ReverseSequence")
-def _ReverseSequenceGrad(op, grad):
+def _ReverseSequenceGrad(op: ops.Operation, grad):
   seq_lengths = op.inputs[1]
   return [
       array_ops.reverse_sequence(
@@ -922,19 +953,19 @@ def _ReverseSequenceGrad(op, grad):
 
 
 @ops.RegisterGradient("Reverse")
-def _ReverseGrad(op, grad):
+def _ReverseGrad(op: ops.Operation, grad):
   reverse_dims = op.inputs[1]
   return gen_array_ops.reverse(grad, reverse_dims), None
 
 
 @ops.RegisterGradient("ReverseV2")
-def _ReverseV2Grad(op, grad):
+def _ReverseV2Grad(op: ops.Operation, grad):
   axis = op.inputs[1]
   return array_ops.reverse_v2(grad, axis), None
 
 
 @ops.RegisterGradient("SpaceToBatch")
-def _SpaceToBatchGrad(op, grad):
+def _SpaceToBatchGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: BatchToSpace.
   block_size = op.get_attr("block_size")
   return [
@@ -943,7 +974,7 @@ def _SpaceToBatchGrad(op, grad):
 
 
 @ops.RegisterGradient("SpaceToBatchND")
-def _SpaceToBatchNDGrad(op, grad):
+def _SpaceToBatchNDGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: BatchToSpaceND.
   return [
       array_ops.batch_to_space_nd(grad, op.inputs[1], op.inputs[2]), None, None
@@ -951,7 +982,7 @@ def _SpaceToBatchNDGrad(op, grad):
 
 
 @ops.RegisterGradient("BatchToSpace")
-def _BatchToSpaceGrad(op, grad):
+def _BatchToSpaceGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: SpaceToBatch.
   block_size = op.get_attr("block_size")
   return [
@@ -960,7 +991,7 @@ def _BatchToSpaceGrad(op, grad):
 
 
 @ops.RegisterGradient("BatchToSpaceND")
-def _BatchToSpaceNDGrad(op, grad):
+def _BatchToSpaceNDGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: SpaceToBatchND.
   return [
       array_ops.space_to_batch_nd(grad, op.inputs[1], op.inputs[2]), None, None
@@ -968,7 +999,7 @@ def _BatchToSpaceNDGrad(op, grad):
 
 
 @ops.RegisterGradient("SpaceToDepth")
-def _SpaceToDepthGrad(op, grad):
+def _SpaceToDepthGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: DepthToSpace.
   block_size = op.get_attr("block_size")
   data_format = op.get_attr("data_format")
@@ -979,7 +1010,7 @@ def _SpaceToDepthGrad(op, grad):
 
 
 @ops.RegisterGradient("DepthToSpace")
-def _DepthToSpaceGrad(op, grad):
+def _DepthToSpaceGrad(op: ops.Operation, grad):
   # Its gradient is the opposite op: SpaceToDepth.
   block_size = op.get_attr("block_size")
   data_format = op.get_attr("data_format")
@@ -993,17 +1024,22 @@ ops.NotDifferentiable("OneHot")
 
 
 @ops.RegisterGradient("MirrorPad")
-def _MirrorPadGrad(op, grad):
+def _MirrorPadGrad(op: ops.Operation, grad):
   mode = op.get_attr("mode")
   return [gen_array_ops.mirror_pad_grad(grad, op.inputs[1], mode=mode), None]
 
 
 @ops.RegisterGradient("MirrorPadGrad")
-def _MirrorPadGradGrad(op, grad):
+def _MirrorPadGradGrad(op: ops.Operation, grad):
   mode = op.get_attr("mode")
   return [gen_array_ops.mirror_pad(grad, op.inputs[1], mode=mode), None]
 
 
+ops.NotDifferentiable("FakeQuantWithMinMaxArgsGradient")
+ops.NotDifferentiable("FakeQuantWithMinMaxVarsGradient")
+ops.NotDifferentiable("FakeQuantWithMinMaxVarsPerChannelGradient")
+
+
 @ops.RegisterGradient("QuantizeAndDequantize")
 def _QuantizeAndDequantizeGrad(_, grad):
   return grad
@@ -1021,7 +1057,7 @@ def _QuantizeAndDequantizeV3Grad(_, grad):
 
 
 @ops.RegisterGradient("ExtractImagePatches")
-def _ExtractImagePatchesGrad(op, grad):
+def _ExtractImagePatchesGrad(op: ops.Operation, grad):  # pylint:disable=missing-function-docstring
   input_bhwc = array_ops.shape(op.inputs[0], out_type=dtypes.int64)
   batch_size, rows_in, cols_in, channels = array_ops_stack.unstack(input_bhwc)
 
@@ -1068,7 +1104,7 @@ def _ExtractImagePatchesGrad(op, grad):
 
 
 @ops.RegisterGradient("ExtractVolumePatches")
-def _ExtractVolumePatchesGrad(op, grad):
+def _ExtractVolumePatchesGrad(op: ops.Operation, grad):  # pylint:disable=missing-function-docstring
   batch_size, planes_in, rows_in, cols_in, channels = [
       dim.value for dim in op.inputs[0].shape.dims
   ]
@@ -1131,14 +1167,14 @@ def _ExtractVolumePatchesGrad(op, grad):
 
 
 @ops.RegisterGradient("ScatterNd")
-def _ScatterNdGrad(op, grad):
+def _ScatterNdGrad(op: ops.Operation, grad):
   indices = op.inputs[0]
   updates_grad = array_ops.gather_nd(grad, indices)
   return [None, updates_grad, None]
 
 
 @ops.RegisterGradient("TensorScatterUpdate")
-def _TensorScatterUpdateGrad(op, grad):
+def _TensorScatterUpdateGrad(op: ops.Operation, grad):
   indices = op.inputs[1]
   updates_grad = array_ops.gather_nd(grad, indices)
   tensor_grad = array_ops.tensor_scatter_update(
@@ -1148,14 +1184,14 @@ def _TensorScatterUpdateGrad(op, grad):
 
 
 @ops.RegisterGradient("TensorScatterAdd")
-def _TensorScatterAddGrad(op, grad):
+def _TensorScatterAddGrad(op: ops.Operation, grad):
   indices = op.inputs[1]
   updates_grad = array_ops.gather_nd(grad, indices)
   tensor_grad = array_ops.identity(grad)
   return [tensor_grad, None, updates_grad]
 
 
-def _TensorScatterMinOrMaxGrad(op, grad):
+def _TensorScatterMinOrMaxGrad(op: ops.Operation, grad):
   """Gradient for TensorScatterMin and TensorScatterMax."""
   indices = op.inputs[1]
   x = op.inputs[0]
@@ -1175,19 +1211,19 @@ def _TensorScatterMinOrMaxGrad(op, grad):
 
 
 @ops.RegisterGradient("TensorScatterMax")
-def _TensorScatterMaxGrad(op, grad):
+def _TensorScatterMaxGrad(op: ops.Operation, grad):
   """Gradient for TensorScatterMax op."""
   return _TensorScatterMinOrMaxGrad(op, grad)
 
 
 @ops.RegisterGradient("TensorScatterMin")
-def _TensorScatterMinGrad(op, grad):
+def _TensorScatterMinGrad(op: ops.Operation, grad):
   """Gradient for TensorScatterMin op."""
   return _TensorScatterMinOrMaxGrad(op, grad)
 
 
 @ops.RegisterGradient("TensorScatterSub")
-def _TensorScatterSubGrad(op, grad):
+def _TensorScatterSubGrad(op: ops.Operation, grad):
   indices = op.inputs[1]
   updates_grad = array_ops.gather_nd(grad, indices)
   tensor_grad = array_ops.identity(grad)
@@ -1195,18 +1231,18 @@ def _TensorScatterSubGrad(op, grad):
 
 
 @ops.RegisterGradient("ScatterNdNonAliasingAdd")
-def _ScatterNdNonAliasingAddGrad(op, grad):
+def _ScatterNdNonAliasingAddGrad(op: ops.Operation, grad):
   indices = op.inputs[1]
   updates_grad = array_ops.gather_nd(grad, indices)
   return [grad, None, updates_grad]
 
 
 @ops.RegisterGradient("BroadcastTo")
-def _BroadcastToGrad(op, grad):
+def _BroadcastToGrad(op: ops.Operation, grad):  # pylint:disable=missing-function-docstring
   input_value = op.inputs[0]
   broadcast_shape = op.inputs[1]
   shape_dtype = dtypes.int32
-  if isinstance(broadcast_shape, ops.Tensor):
+  if isinstance(broadcast_shape, tensor.Tensor):
     shape_dtype = broadcast_shape.dtype
 
   input_value_shape = array_ops.shape(input_value, out_type=shape_dtype)
